from datetime import timedelta
from typing import Optional

import pandas as pd
import plotly.express as px

from sampo.utilities.visualization.base import VisualizationMode, visualize


def schedule_gant_chart_fig(schedule_dataframe: pd.DataFrame,
                            visualization: VisualizationMode,
                            fig_file_name: Optional[str] = None):
    """
    Creates and saves a gant chart of the scheduled tasks to the specified path.
    :param fig_file_name:
    :param visualization:
    :param schedule_dataframe: Pandas DataFrame with the information about schedule
    """
    fig = px.timeline(schedule_dataframe, x_start='start', x_end='finish', y='idx', hover_name='task_name',
                      color=schedule_dataframe.loc[:, 'contractor'],
                      hover_data=['workers', 'contractor'],
                      title=f"{'Project tasks - Gant chart'}",
                      category_orders={'idx': list(schedule_dataframe.idx)},
                      text='task_name')

    fig.update_traces(textposition='outside')

    fig.update_yaxes(showticklabels=False, title_text='Project tasks',
                     type="category")

    fig.update_xaxes(type="date",
                     range=[schedule_dataframe.loc[:, 'start'].min() - timedelta(days=2),
                            schedule_dataframe.loc[:, 'finish'].max() + timedelta(days=40)],
                     title_text='Date')

    fig.update_layout(autosize=True, font_size=12)

    return visualize(fig, mode=visualization, file_name=fig_file_name)



from typing import List, Dict, Optional, Tuple

from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.scheduler.heft.time_computaion import work_priority, calculate_working_time_cascade
from sampo.schemas.graph import GraphNode, WorkGraph


def ford_bellman(wg: WorkGraph, weights: Dict[GraphNode, float]) -> Dict[GraphNode, float]:
    path_weights: Dict[GraphNode, float] = {node: 0 for node in wg.nodes}
    # cache graph edges
    edges: List[Tuple[GraphNode, GraphNode, float]] = sorted([(finish, start, weights[finish])
                                                              for start in wg.nodes
                                                              for finish in start.parents],
                                                             key=lambda x: (x[0].id, x[1].id))
    # for changes heuristic
    changed = False
    # run standard ford-bellman on reversed edges
    # optimize dict access to finish weight
    for i in range(wg.vertex_count):
        cur_finish = edges[0][0]
        cur_finish_weight = path_weights[cur_finish]
        # relax on edges
        for finish, start, weight in edges:
            # we are running through the equality class by finish node
            # so if it changes renew the parameters of current equality class
            if cur_finish.id != finish.id:
                path_weights[cur_finish] = cur_finish_weight
                cur_finish = finish
                cur_finish_weight = path_weights[cur_finish]
            new_weight = path_weights[start] + weight
            if new_weight < cur_finish_weight:
                cur_finish_weight = new_weight
                changed = True
        # if we were done completely nothing with actual minimum weights, the algorithm ends
        if not changed:
            break
        # we go here if changed = True
        # so the last equality class weight can be changed, save it
        path_weights[cur_finish] = cur_finish_weight
        # next iteration should start without change info from previous
        changed = False

    return path_weights


def prioritization(wg: WorkGraph, work_estimator: Optional[WorkTimeEstimator] = None) -> List[GraphNode]:
    # inverse weights
    weights = {node: -work_priority(node, calculate_working_time_cascade, work_estimator)
               for node in wg.nodes}

    path_weights = ford_bellman(wg, weights)

    ordered_nodes = [i[0] for i in sorted(path_weights.items(), key=lambda x: (x[1], x[0].id), reverse=True)
                     if not i[0].is_inseparable_son()]

    return ordered_nodes



from typing import List, Callable

import numpy as np

from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.resource.coordinate_descent import CoordinateDescentResourceOptimizer
from sampo.schemas.contractor import WorkerContractorPool
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.utilities.base_opt import dichotomy_int


class FullScanResourceOptimizer(ResourceOptimizer):

    _coordinate_descent_optimizer = CoordinateDescentResourceOptimizer(dichotomy_int)

    def optimize_resources(self,
                           worker_pool: WorkerContractorPool,
                           worker_team: List[Worker],
                           optimize_array: np.ndarray,
                           down_border: np.ndarray,
                           up_border: np.ndarray,
                           get_finish_time: Callable[[List[Worker]], Time]):

        # TODO Handle optimize_array
        def fitness(worker_count: np.ndarray):
            for worker_ind in range(len(worker_team)):
                worker_team[worker_ind].count = worker_count[worker_ind]
            return get_finish_time(worker_team)

        cur = down_border.copy()
        cur_ft = fitness(cur)

        # Trying to +1 to all workers
        while (up_border > cur).all():
            cur += 1
            next_ft = fitness(cur)
            if next_ft >= cur_ft:
                cur -= 1
                break

        # Insert cur into coordinate-descent optimizer as down_border
        self._coordinate_descent_optimizer.optimize_resources(worker_pool,
                                                              worker_team,
                                                              optimize_array,
                                                              cur,
                                                              up_border,
                                                              get_finish_time)



from datetime import datetime
from functools import partial, lru_cache
from typing import Tuple, List, Dict, Iterable, Union, Set

from pandas import DataFrame

from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.serializable import JSONSerializable, T
from sampo.schemas.time import Time
from sampo.schemas.works import WorkUnit
from sampo.utilities.datetime import add_time_delta
from sampo.utilities.schedule import fix_split_tasks

ResourceSchedule = Dict[str, List[Tuple[Time, Time]]]
ScheduleWorkDict = Dict[str, ScheduledWork]


# TODO: Rebase object onto ScheduleWorkDict and ordered ScheduledWork list
class Schedule(JSONSerializable['Schedule']):
    """
    Represents work schedule. Is a wrapper around DataFrame with specific structure.
    """
    _schedule: DataFrame

    _data_columns: List[str] = ['idx', 'task_id', 'task_name', 'contractor', 'volume',
                                'measurement', 'successors', 'start',
                                'finish', 'duration', 'workers']
    _scheduled_work_column: str = 'scheduled_work_object'

    _columns: List[str] = _data_columns + [_scheduled_work_column]

    @property
    def full_schedule_df(self) -> DataFrame:
        """
        The full schedule DataFrame with all works, data columns and a distinct column for ScheduledWork objects
        :return: Full schedule DataFrame.
        """
        return self._schedule

    @property
    def pure_schedule_df(self) -> DataFrame:
        """
        Schedule DataFrame without service units and containing only original columns (stored in _data_columns field).
        :return: Pure schedule DataFrame.
        """
        return self._schedule[~self._schedule.apply(
            lambda row: row[self._scheduled_work_column].work_unit.is_service_unit,
            axis=1
        )][self._data_columns]

    @property
    def works(self) -> Iterable[ScheduledWork]:
        """
        Enumerates ScheduledWorks in the Schedule.
        :return: Iterable collection of all the scheduled works.
        """
        return self._schedule.scheduled_work_object

    @property
    def to_schedule_work_dict(self) -> ScheduleWorkDict:
        """
        Builds a ScheduleWorkDict from the Schedule.
        :return: ScheduleWorkDict with all the scheduled works.
        """
        return {r['task_id']: r['scheduled_work_object'] for _, r in self._schedule.iterrows()}

    @property
    def execution_time(self) -> Time:
        """
        Calculates total schedule execution time.
        :return: Finish time of the last work.
        """
        return self._schedule.iloc[-1].finish

    def __init__(self, schedule: DataFrame) -> None:
        """
        Initializes new `Schedule` object as a wrapper around `DataFrame` with specific structure.
        Don't use manually. Create Schedule `objects` via `from_scheduled_works` factory method.
        :param schedule: Prepared schedule `DataFrame`.
        """
        self._schedule = schedule

    # [SECTION] JSONSerializable overrides
    def _serialize(self) -> T:
        # Method described in base class
        return {
            'works': [sw._serialize() for sw in self._schedule.scheduled_work_object]
        }

    @classmethod
    def _deserialize(cls, dict_representation: T) -> 'Schedule':
        # Method described in base class
        dict_representation['works'] = [ScheduledWork._deserialize(sw) for sw in dict_representation['works']]
        return Schedule.from_scheduled_works(**dict_representation)

    @lru_cache
    def merged_stages_datetime_df(self, offset: Union[datetime, str]) -> DataFrame:
        """
        Merges split stages of same works after lag optimization and returns schedule DataFrame shifted to start.
        :param offset: Start of schedule, to add as an offset.
        :return: Shifted schedule DataFrame with merged tasks.
        """
        result = fix_split_tasks(self.offset_schedule(offset))
        return result

    def offset_schedule(self, offset: Union[datetime, str]) -> DataFrame:
        """
        Returns full schedule object with `start` and `finish` columns pushed by date in `offset` argument.
        :param offset: Start of schedule, to add as an offset.
        :return: Shifted schedule DataFrame.
        """
        r = self._schedule.loc[:, :]
        r['start_offset'] = r['start'].apply(partial(add_time_delta, offset))
        r['finish_offset'] = r['finish'].apply(partial(add_time_delta, offset))
        r = r.rename({'start': 'start_', 'finish': 'finish_',
                      'start_offset': 'start', 'finish_offset': 'finish'}, axis=1) \
            .drop(['start_', 'finish_'], axis=1)
        return r

    @staticmethod
    def from_scheduled_works(works: Iterable[ScheduledWork],
                             wg: WorkGraph = None) \
            -> 'Schedule':
        """
        Factory method to create a Schedule object from list of Schedule works and additional info
        :param wg: Work graph.
        :param works: Iterable collection of ScheduledWork's.
        :return: Schedule.
        """
        ordered_task_ids = order_nodes_by_start_time(works, wg) if wg else None

        def info(work_unit: WorkUnit) -> Tuple[float, str, List[Tuple[str, str]]]:
            if not wg:
                return 0, "", []
            # noinspection PyTypeChecker
            return work_unit.volume, work_unit.volume_type, \
                   [(edge.finish.id, edge.type.value) for edge in wg[work_unit.id].edges_from]

        def sed(t1, t2) -> tuple:
            """
            Sorts times and calculates difference.
            :param t1: time 1.
            :param t2: time 2.
            :return: Tuple: start, end, duration.
            """
            s, e = tuple(sorted((t1, t2)))
            return s, e, e - s

        df = [(i,  # idx
               w.work_unit.id,  # task_id
               w.work_unit.name,  # task_name
               w.contractor,  # contractor info
               *info(w.work_unit),  # volume, measurement, successors
               *sed(*(t.value for t in w.start_end_time)),  # start, end, duration
               repr(dict((i.name, i.count) for i in w.workers)),  # workers
               w  # full ScheduledWork info
               ) for i, w in enumerate(works)]
        df = DataFrame.from_records(df, columns=Schedule._columns)

        df = df.set_index('idx')

        if ordered_task_ids:
            df.task_id = df.task_id.astype('category')
            df.task_id = df.task_id.cat.set_categories(ordered_task_ids)
            df = df.sort_values(['task_id'])
            df.task_id = df.task_id.astype(str)

        df = df.reindex(columns=Schedule._columns)
        df = df.reset_index(drop=True)

        return Schedule(df)


def order_nodes_by_start_time(works: Iterable[ScheduledWork], wg: WorkGraph) -> List[str]:
    """
    Makes ScheduledWorks' ordering that satisfies:
    1. Ascending order by start time
    2. Toposort
    :param works:
    :param wg:
    :return:
    """
    res = []
    order_by_start_time = [(item.start_time, item.work_unit.id) for item in
                           sorted(works, key=lambda item: item.start_time)]

    cur_time = 0
    cur_class: Set[GraphNode] = set()
    for start_time, work in order_by_start_time:
        node = wg[work]
        if len(cur_class) == 0:
            cur_time = start_time
        if start_time == cur_time:
            cur_class.add(node)
            continue
        # TODO Perform real toposort
        cur_not_added: Set[GraphNode] = set(cur_class)
        while len(cur_not_added) > 0:
            for cur_node in cur_class:
                if any([parent_node in cur_not_added for parent_node in cur_node.parents]):
                    continue  # we add this node later
                res.append(cur_node.id)
                cur_not_added.remove(cur_node)
            cur_class = set(cur_not_added)
        cur_time = start_time
        cur_class = {node}

    cur_not_added: Set[GraphNode] = set(cur_class)
    while len(cur_not_added) > 0:
        for cur_node in cur_class:
            if any([parent_node in cur_not_added for parent_node in cur_node.parents]):
                continue  # we add this node later
            res.append(cur_node.id)
            cur_not_added.remove(cur_node)
        cur_class = set(cur_not_added)

    return res



from functools import partial
from io import StringIO
from typing import Union, Callable

import numpy as np
import pandas as pd

CUSTOM_FIELD_SERIALIZER = '_serializer_for_fields'
CUSTOM_FIELD_DESERIALIZER = '_deserializer_for_fields'
CUSTOM_TYPE_SERIALIZER = '_serializer_for_types'
CUSTOM_TYPE_DESERIALIZER = '_deserializer_for_types'


def custom_serializer(type_or_field: Union[type, str], deserializer: bool = False):
    """
    Meta-decorator for marking custom serializers or deserializers methods.<br/>
    This decorator can stack with other serializer/deserializer decorators.
    :param type_or_field: Name (str) of field or type (type) of fields, which will be serialized with this serializer in
    current class. If type should be presented in str representation, consider using custom_type_serializer or
    custom_type_deserializer decorators.
    :param deserializer:
    If True, the decorated function will be considered as a custom deserializer for type_or_field type or field<br/>
    If None, deserializer should be decorated separately with custom_serializer or custom_field_deserializer or
    custom_type_deserializer decorator
    :return:
    """
    type_of_entity = type(type_or_field)
    if type_of_entity is type:
        # if callable(deserializer):
        #     _decorate_serializer(func=deserializer,
        #                          collection_name=CUSTOM_TYPE_DESERIALIZER,
        #                          new_element=type_or_field)
        # elif deserializer:
        if deserializer:
            return custom_type_deserializer(type_or_field)
        return custom_type_serializer(type_or_field)

    elif type_of_entity is str:
        # if callable(deserializer):
        #     _decorate_serializer(func=deserializer,
        #                          collection_name=CUSTOM_FIELD_DESERIALIZER,
        #                          new_element=type_or_field)
        # elif deserializer:
        if deserializer:
            return custom_field_deserializer(type_or_field)
        return custom_field_serializer(type_or_field)

    raise TypeError(f'Unexpected type of param type_or_field: {type(type_or_field)} instead of Union[type, str]')


def custom_field_serializer(field_name: str):
    return partial(_decorate_serializer,
                   collection_name=CUSTOM_FIELD_SERIALIZER,
                   new_element=field_name)


def custom_field_deserializer(field_name: str):
    return partial(_decorate_serializer,
                   collection_name=CUSTOM_FIELD_DESERIALIZER,
                   new_element=field_name)


def custom_type_serializer(__type: type or str):
    return partial(_decorate_serializer,
                   collection_name=CUSTOM_TYPE_SERIALIZER,
                   new_element=__type)


def custom_type_deserializer(__type: type or str):
    return partial(_decorate_serializer,
                   collection_name=CUSTOM_TYPE_DESERIALIZER,
                   new_element=__type)


def _decorate_serializer(func, collection_name, new_element):
    if not hasattr(func, collection_name):
        setattr(func, collection_name, [])
    getattr(func, collection_name).append(new_element)
    return func


def default_ndarray_serializer(array: np.ndarray):
    return array.tolist()


def default_ndarray_deserializer(__list: list):
    return np.array(__list)


def default_dataframe_serializer(df: pd.DataFrame):
    return df.to_csv(encoding='utf-8')


def default_dataframe_deserializer(str_repr: str):
    return pd.read_csv(StringIO(str_repr), encoding='utf-8')


def default_np_int_serializer(n):
    return int(n)


def default_np_int_deserializer(n):
    return np.int32(n)




from sampo.schemas.serializable import AutoJSONSerializable

TIME_INF = 2_000_000


# TODO Consider converting to StrSerializable
# TODO: describe the class (description, parameters/)
class Time(AutoJSONSerializable['Time']):
    value: int
    
    # TODO: describe the function (description, parameters, return type)
    def __init__(self, value: int = 0):
        self.value = 0
        self.set_time(value)

    # TODO: describe the function (description, return type)
    @staticmethod
    def inf():
        return Time(TIME_INF)

    # Copying body between left- and right-associative functions
    # is here to avoid functional call overhead
    # TODO: describe the function (description, parameters, return type)
    def __add__(self, other: 'Time' or int) -> 'Time':
        return Time(self.value + (other.value if isinstance(other, Time) else other))

    # TODO: describe the function (description, parameters, return type)
    def __radd__(self, other: 'Time' or int) -> 'Time':
        return Time(self.value + (other.value if isinstance(other, Time) else other))

    # TODO: describe the function (description, parameters, return type)
    def __sub__(self, other: 'Time' or int) -> 'Time':
        return Time(self.value - (other.value if isinstance(other, Time) else other))

    # TODO: describe the function (description, parameters, return type)
    def __rsub__(self, other: 'Time' or int) -> 'Time':
        return Time((other.value if isinstance(other, Time) else other) - self.value)

    # TODO: describe the function (description, parameters, return type)
    def __mul__(self, other: 'Time' or int) -> 'Time':
        return Time(self.value * (other.value if isinstance(other, Time) else other))

    # TODO: describe the function (description, parameters, return type)
    def __rmul__(self, other: 'Time' or int) -> 'Time':
        return Time(self.value * (other.value if isinstance(other, Time) else other))

    # TODO: describe the function (description, parameters, return type)
    def __floordiv__(self, other: 'Time' or int) -> 'Time':
        return Time(self.value // (other.value if isinstance(other, Time) else other))

    # TODO: describe the function (description, parameters, return type)
    def __truediv__(self, other: 'Time' or int) -> float:
        return self.value / (other.value if isinstance(other, Time) else other)

    # TODO: describe the function (description, parameters, return type)
    def __lt__(self, other):
        return self.value < other

    # TODO: describe the function (description, parameters, return type)
    def __le__(self, other):
        return self.value <= other

    # TODO: describe the function (description, parameters, return type)
    def __gt__(self, other):
        return self.value > other

    # TODO: describe the function (description, parameters, return type)
    def __ge__(self, other):
        return self.value >= other

    # TODO: describe the function (description, parameters, return type)
    def __eq__(self, other):
        return self.value == other

    # TODO: describe the function (description, return type)
    def __bool__(self):
        return self.value != 0

    # TODO: describe the function (description, return type)
    def __int__(self) -> int:
        return self.value

    # TODO: describe the function (description, return type)
    def __str__(self) -> str:
        return str(self.value)

    # TODO: describe the function (description, return type)
    def __repr__(self) -> str:
        return str(self.value)

    # TODO: describe the function (description, return type)
    def __unicode__(self) -> str:
        return str(self.value)

    # TODO: describe the function (description, return type)
    def __hash__(self) -> int:
        return hash(self.value)

    # TODO: describe the function (description, return type)
    def set_time(self, value: int):
        value = int(value)
        if value > TIME_INF:
            value = TIME_INF
        elif value < -TIME_INF:
            value = -TIME_INF
        self.value = value

    # TODO: describe the function (description, return type)
    def is_inf(self) -> bool:
        return abs(self.value) == TIME_INF



from sampo.schemas.schedule import Schedule


def schedule_cost(schedule: Schedule) -> float:
    cost: float = 0

    for work in schedule.works:
        for worker in work.workers:
            cost += worker.get_cost() * work.duration.value
    return cost



from random import Random
from typing import Optional, Dict, List

from sampo.generator.config.gen_counts import WORKER_PROPORTIONS
from sampo.schemas.contractor import Contractor
from sampo.schemas.interval import IntervalGaussian
from sampo.schemas.resources import Worker
from sampo.schemas.utils import uuid_str


def get_stochastic_counts(pack_count: float, sigma_scaler: float, proportions: Optional[Dict[str, float]],
                          available_types: Optional[List] = None, rand: Optional[Random] = None) -> Dict[str, int]:
    available_types = available_types or list(proportions.keys())
    counts = {name: prop * pack_count for name, prop in proportions.items() if name in available_types}
    stochastic_counts = {
        name: IntervalGaussian(count, sigma_scaler * count / 2, proportions[name], count).rand_int(rand)
        for name, count in counts.items()
    }
    return stochastic_counts


def dict_subtract(d: Dict[str, float], subtractor: float) -> Dict[str, float]:
    for name in d.keys():
        d[name] -= subtractor
    return d


def get_contractor_with_equal_proportions(number_of_workers_in_contractors: int or List[int],
                                          number_of_contractors: int = 1) \
        -> List[Contractor]:
    """
    Generates a contractors list of specified length with specified capacities
    :param number_of_workers_in_contractors: How many workers of all each contractor contains in itself.
    One int for all or List[int] for each contractor. If List, its length should be equal to number_of_contractors.
    :param number_of_contractors: Number of generated contractors.
    :return: List with contractors
    """
    assert isinstance(number_of_workers_in_contractors, int) \
           or len(number_of_workers_in_contractors) == number_of_contractors, \
        'Wrong arguments. Length of number_of_workers_in_contractors should equal to number_of_contractors,' + \
        ' or its value should be int,'
    if isinstance(number_of_workers_in_contractors, int):
        number_of_workers_in_contractors = [number_of_workers_in_contractors] * number_of_contractors

    return [get_contractor(workers_count) for workers_count in number_of_workers_in_contractors]


def get_contractor(pack_worker_count: float,
                   sigma_scaler: Optional[float] = 0.1,
                   index: int = 0,
                   worker_proportions: Optional[Dict[str, float]] = WORKER_PROPORTIONS,
                   available_worker_types: Optional[List] = None, rand: Optional[Random] = None) -> Contractor:
    contractor_id = uuid_str(rand)

    worker_counts = get_stochastic_counts(pack_worker_count, sigma_scaler, worker_proportions,
                                          available_worker_types, rand)

    workers = dict()
    for name in worker_counts.keys():
        # print(counts_classes, worker_counts[name])
        # counts_classes[START_BASIC_CLASS] += worker_proportions[name]
        # counts_classes[START_BASIC_CLASS] += max((worker_counts[name] - sum(counts_classes)), 0)
        workers[name] = Worker(uuid_str(rand), name,
                               contractor_id=contractor_id,
                               count=worker_counts[name])

    return Contractor(contractor_id, f"Подрядчик {index}", workers, {})
